import javax.sql.DataSource

import grails.util.GrailsUtil

import org.apache.commons.dbcp.BasicDataSource

import org.codehaus.groovy.grails.commons.ApplicationHolder as AH
import org.codehaus.groovy.grails.plugins.orm.hibernate.HibernateGrailsPlugin
import org.codehaus.groovy.grails.orm.hibernate.ConfigurableLocalSessionFactoryBean
import org.codehaus.groovy.grails.orm.hibernate.support.GrailsOpenSessionInViewInterceptor
import org.codehaus.groovy.grails.orm.hibernate.support.HibernatePersistenceContextInterceptor
import org.codehaus.groovy.grails.orm.hibernate.support.SpringLobHandlerDetectorFactoryBean
import org.codehaus.groovy.grails.web.servlet.mvc.GrailsUrlHandlerMapping

import org.springframework.beans.factory.config.PropertiesFactoryBean
import org.springframework.jdbc.datasource.DriverManagerDataSource
import org.springframework.jndi.JndiObjectFactoryBean
import org.springframework.orm.hibernate3.HibernateAccessor
import org.springframework.orm.hibernate3.HibernateTransactionManager
import org.springframework.transaction.interceptor.TransactionProxyFactoryBean

import com.burtbeckwith.grails.plugin.datasources.DatasourcesBuilder
import com.burtbeckwith.grails.plugin.datasources.DatasourcesUtils
import com.burtbeckwith.grails.plugin.datasources.FakeApplicationContext
import com.burtbeckwith.grails.plugin.datasources.FakeGrailsApplication
import com.burtbeckwith.grails.plugin.datasources.ReadOnlyDriverManagerDataSource
import com.burtbeckwith.grails.plugin.datasources.model.DatasourceConfig

class DatasourcesGrailsPlugin {

	def version = 0.1
	def dependsOn = [hibernate: grails.util.GrailsUtil.grailsVersion]
	def loadAfter = ['hibernate', 'services']
	def watchedResources = 'file:./grails-app/conf/Datasources.groovy'
	def author = 'Burt Beckwith'
	def authorEmail = 'beckwithb@studentsonly.com'
	def title = 'Grails Datasources Plugin.'
	def description = 'Supports defining multiple data sources per application.'
	def documentation = 'http://burtbeckwith.com/blog/?p=70'

	def doWithSpring = {

		def dataSources = loadConfig()
		if (!dataSources) {
			return
		}

		def handlerInterceptors = []
		if (manager?.hasGrailsPlugin('controllers')) {
			handlerInterceptors = [ref('localeChangeInterceptor'), ref('openSessionInViewInterceptor')]
		}

		def application = AH.application

		for (ds in dataSources) {

			if (!ds.environments.contains(GrailsUtil.environment)) {
				return
			}

			String name = ds.name
			def dsClass = ds.jndiName ? JndiObjectFactoryBean :
				ds.pooled ? BasicDataSource :
				ds.readOnly ? ReadOnlyDriverManagerDataSource : DriverManagerDataSource

			def dsBean = "dataSource_$name"(dsClass) {
				if (ds.jndiName) {
					jndiName = ds.jndiName
					expectedType = DataSource
				}
				else {
					if (ds.pooled) {
						defaultReadOnly = ds.readOnly
					}
					driverClassName = ds.driverClassName
					url = ds.url
					if (ds.username) {
						username = ds.username
					}
					if (ds.password) {
						password = ds.password
					}
				}
			}
			if (ds.pooled) {
				dsBean.destroyMethod = 'close'
			}

			"hibernateProperties_$name"(PropertiesFactoryBean) { bean ->
				bean.scope = 'prototype'
				properties = ds.hibernateProperties
			}
			"lobHandlerDetector_$name"(SpringLobHandlerDetectorFactoryBean) {
				dataSource = ref("dataSource_$name")
			}

			def domainClasses = findDataSourceDomainClasses(ds, application)
			"sessionFactory_$name"(ConfigurableLocalSessionFactoryBean) {
				dataSource = ref("dataSource_$name")
				hibernateProperties = ref("hibernateProperties_$name")
				grailsApplication = new FakeGrailsApplication(domainClasses: domainClasses)
				lobHandler = ref("lobHandlerDetector_$name")
				entityInterceptor = ref('eventTriggeringInterceptor')
			}

			"transactionManager_$name"(HibernateTransactionManager) {
				sessionFactory = ref("sessionFactory_$name")
			}

			"persistenceInterceptor_$name"(HibernatePersistenceContextInterceptor) {
				sessionFactory = ref("sessionFactory_$name")
			}

			if (manager?.hasGrailsPlugin('controllers')) {
				"openSessionInViewInterceptor_$name"(GrailsOpenSessionInViewInterceptor) {
					flushMode = ds.readOnly ? HibernateAccessor.FLUSH_NEVER : HibernateAccessor.FLUSH_AUTO
					sessionFactory = ref("sessionFactory_$name")
				}

				handlerInterceptors << ref("openSessionInViewInterceptor_$name")
			}

			// configure transactional services
			if (ds.services) {
				application.serviceClasses.each { serviceClass ->
					if (serviceClass.transactional &&
							ds.services.contains(serviceClass.propertyName - 'Service')) {

						def scope = serviceClass.getPropertyValue("scope")
						def props = new Properties()
						String attributes = 'PROPAGATION_REQUIRED'
						if (ds.readOnly) {
							attributes += ',readOnly'
						}
						props."*" = attributes
						"${serviceClass.propertyName}"(TransactionProxyFactoryBean) { bean ->
					    	if (scope) bean.scope = scope
					    	target = { innerBean ->
					    		innerBean.factoryBean = "${serviceClass.fullName}ServiceClass"
					    		innerBean.factoryMethod = "newInstance"
					    		innerBean.autowire = "byName"
					    		if (scope) innerBean.scope = scope
					    	}
					    	proxyTargetClass = true
					    	transactionAttributes = props
					    	transactionManager = ref("transactionManager_$name")
						}
					}
				}
			}
		}

		// redefine with the extra OSIV interceptors
		grailsUrlHandlerMapping(GrailsUrlHandlerMapping) {
			interceptors = handlerInterceptors
			mappings = ref('grailsUrlMappings')
		}
	}

	private def findDataSourceDomainClasses(ds, application) {
		def classNames = ds.domainClasses as Set
		def domainClasses = application.domainClasses.findAll { dc ->
			// need to check name in case of app reload
			classNames.contains(dc.clazz.name)
		}
		return domainClasses
	}

	def doWithDynamicMethods = { ctx ->

		def application = AH.application
		def dataSources = loadConfig()
		if (!dataSources) {
			return
		}

		DatasourcesUtils.applicationContext = ctx
		Closure c = new HibernateGrailsPlugin().doWithDynamicMethods

		for (ds in dataSources) {

			if (!ds.environments.contains(GrailsUtil.environment)) {
				return
			}

			String name = ds.name
			def domainClasses = findDataSourceDomainClasses(ds, application)
			def fakeApplication = new FakeGrailsApplication(domainClasses: domainClasses)

			def closure = c.clone()
			closure.delegate = [application: fakeApplication]
			closure.call new FakeApplicationContext(
					transactionManager: ctx.getBean("transactionManager_$name"),
					sessionFactory: ctx.getBean("sessionFactory_$name"))
		}
	}

	def doWithApplicationContext = { applicationContext ->
		// nothing to do
	}

	def doWithWebDescriptor = { xml ->
		// nothing to do
	}

	def onChange = { event ->
		// nothing to do
	}

	def onConfigChange = { event ->
		// nothing to do
	}

	def onShutdown = { event ->
		// todo: HSQLDB shutdown like in DataSourceGrailsPlugin
	}

	private def loadConfig() {

	    def script = AH.application.classLoader.loadClass('Datasources').newInstance()
	    script.run()

	    def builder = new DatasourcesBuilder()
	    def datasources = script.datasources
	    datasources.delegate = builder
	    datasources()

		return builder.dataSources
	}
}
